/*
### Written by Torsten Moeller
### sometime in March 1998
### let me know, if you find any errors (moeller@cis.ohio-state.edu)
*/

#include "tmFilters.h"
#include "tmFilters_raw.h"
#include <stdio.h>
#include <math.h>

//////////////////////////////////////////////////////////////////////////
// instantiation
//////////////////////////////////////////////////////////////////////////

// the default filter will be C^{-1} continuous 1EF interpolation filter.
tmFilter::tmFilter()
{
  a = 0; // open parameter
  pickFilter(0, -1, 1);
  /*
  D = 0; // derivative
  C = -1; // continuity
  A = 1; // accuracy
  filter = ; // the actual filter
  */
}

tmFilter::tmFilter(int derivative, int continuity, int accuracy)
{
  a = 0; // open parameter
  pickFilter(derivative, continuity, accuracy);
}

//////////////////////////////////////////////////////////////////////////
// utility functions
//////////////////////////////////////////////////////////////////////////

// We can select which filter to use
void tmFilter::setFunc(int derivative, int continuity, int accuracy)
{
  pickFilter(derivative, continuity, accuracy);
}

// specifies the open parameter that some (not all!) filters have
void tmFilter::setA(float a)
{
  this->a = a;
}

// what is the (non-zero) length of the filter?
float tmFilter::getSize(void)
{
  return size;
}

// We can output the filter - we just need to specify the sampling distance
void tmFilter::print(float T)
{
  int i = 0;

  printf("Filter:\n%i - derivative\n", D);
  printf("%i - continuity\n", C);
  printf("%i - accuracy\n", A);
  printf("%f - open parameter\n", a);

  while (i*T < size) {
    printf("%f %f\n", i*T, m_filter(a, i*T));
    i++;
  }
}

// this is the main thing of this class - it returns filter values
float tmFilter::operator[](float x)
{
  return m_filter(a, x);
}

// this is the main thing of this class - it returns filter values
float tmFilter::filter(float x)
{
  return m_filter(a, x);
}

//////////////////////////////////////////////////////////////////////////
// picking the right filter function
//////////////////////////////////////////////////////////////////////////
// We can select which filter to use
void tmFilter::pickFilter(int derivative, int continuity, int accuracy)
{
  int error = 0;

  D = derivative;
  C = continuity;
  A = accuracy;

  ///////////////////////////////
  // interpolation filter      //
  ///////////////////////////////
  if (D == 0) {
    switch (C) {
    case -1: // non-continuous
      switch (A) {
      case 1:
	m_filter = d0_cn_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d0_cn_2ef;
	size = 2;
	break;
      case 3:
	m_filter = d0_cn_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d0_cn_4ef;
	size = 4;
	break;
      default:
	error = 1;
      }
      break;
    case 0: // C^0
      switch (A) {
      case 1:
	m_filter = d0_c0_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d0_c0_2ef;
	size = 2;
	break;
      case 3:
	m_filter = d0_c0_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d0_c0_4ef;
	size = 4;
	break;
      default:
	error = 1;
      }
      break;
    case 1: // C^1
      switch (A) {
      case 1:
	m_filter = d0_c1_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d0_c1_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d0_c1_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d0_c1_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 2: // C^2
      switch (A) {
      case 1:
	m_filter = d0_c2_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d0_c2_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d0_c2_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d0_c2_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 3: // C^3
      switch (A) {
      case 1:
	m_filter = d0_c3_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d0_c3_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d0_c3_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d0_c3_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    default:
      error = 1;
    }

  ///////////////////////////////
  // first derivative m_filter //
  ///////////////////////////////
  } else if (D == 1) {
    switch (C) {
    case -1: // non-continuous
      switch (A) {
      case 1:
	m_filter = d1_cn_1ef;
	size = 2;
	break;
      case 2:
	m_filter = d1_cn_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d1_cn_3ef;
	size = 4;
	break;
      case 4:
	m_filter = d1_cn_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 0: // C^0
      switch (A) {
      case 1:
	m_filter = d1_c0_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d1_c0_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d1_c0_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d1_c0_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 1: // C^1
      switch (A) {
      case 1:
	m_filter = d1_c1_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d1_c1_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d1_c1_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d1_c1_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 2: // C^2
      switch (A) {
      case 1:
	m_filter = d1_c2_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d1_c2_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d1_c2_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d1_c2_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 3: // C^3
      switch (A) {
      case 1:
	m_filter = d1_c3_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d1_c3_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d1_c3_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d1_c3_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    default:
      error = 1;
    }



  ///////////////////////////////
  // second derivative filter  //
  ///////////////////////////////
  } else if (D == 2) {
    switch (C) {
    case -1: // non-continuous
      switch (A) {
      case 1:
	m_filter = d2_cn_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d2_cn_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d2_cn_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d2_cn_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 0: // C^0
      switch (A) {
      case 1:
	m_filter = d2_c0_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d2_c0_2ef;
	size = 4;
	break;
      case 3:
	m_filter = d2_c0_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d2_c0_4ef;
	size = 6;
	break;
      default:
	error = 1;
      }
      break;
    case 1: // C^1
      switch (A) {
      case 1:
	m_filter = d2_c1_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d2_c1_2ef;
	size = 6;
	break;
      case 3:
	m_filter = d2_c1_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d2_c1_4ef;
	size = 8;
	break;
      default:
	error = 1;
      }
      break;
    case 2: // C^2
      switch (A) {
      case 1:
	m_filter = d2_c2_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d2_c2_2ef;
	size = 6;
	break;
      case 3:
	m_filter = d2_c2_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d2_c2_4ef;
	size = 8;
	break;
      default:
	error = 1;
      }
      break;
    case 3: // C^3
      switch (A) {
      case 1:
	m_filter = d2_c3_1ef;
	size = 4;
	break;
      case 2:
	m_filter = d2_c3_2ef;
	size = 6;
	break;
      case 3:
	m_filter = d2_c3_3ef;
	size = 6;
	break;
      case 4:
	m_filter = d2_c3_4ef;
	size = 8;
	break;
      default:
	error = 1;
      }
      break;
    default:
      error = 1;
    }

  } else
    error = 1;

  if (error)
    printf("I couldn't find a filter of (D, C, A) = (%i, %i, %i)", D, C, A);
}
